import sys
import json
import typing as T
from . import meson_exe

# This script is used by run_unittests.py to verify we don't load too many
# modules when executing a wrapped command.
def run(args: T.List[str]) -> int:
    meson_exe.run(args)
    print(json.dumps(list(sys.modules.keys())))
    return 0
